import json
from pathlib import Path

from ..schema import FFProbeInfo
from .hotfix_webp import hotfix_animate_webp
from .loader import from_dict
from .shell import call


def ffprobe(input_url: str | Path) -> FFProbeInfo:
    """
    Get media information using FFprobe.

    Args:
        input_url: the URI of the media file

    Returns:
        the media information

    Raises:
        FfmpegMediaTypeError: If the FFprobe command fails.
    """

    input_url = hotfix_animate_webp(input_url)

    # Construct the FFprobe command with JSON output format
    ffprobe_cmd = ["ffprobe"] + [
        "-v",
        "error",
        "-show_format",
        "-show_streams",
        "-of",
        "json",
        str(input_url),
    ]

    # Execute the FFprobe command and capture the output
    output = call(ffprobe_cmd)
    probe_info = from_dict(FFProbeInfo, json.loads(output))

    return probe_info



import json
import subprocess
from functools import lru_cache
from pathlib import Path
from urllib.request import urlretrieve

from ..schema import FFProbeInfo
from .loader import from_dict
from .shell import call, create_temp_filename


def _is_url(uri: str | Path) -> bool:
    """
    Check if the URI is a URL.

    Args:
        uri: The URI to check.

    Returns:
        True if the URI is a URL, False otherwise.
    """
    return isinstance(uri, str) and uri.startswith(("http://", "https://"))


def _ensure_downloaded(uri: str) -> str:
    """
    Download a file from a URL.

    Args:
        uri: The URL of the file to download.

    Returns:
        The path to the downloaded file.
    """
    if not _is_url(uri):
        return uri

    # Download the file
    file_path = create_temp_filename(".webp")
    urlretrieve(uri, file_path)

    return file_path


@lru_cache()
def check_webpmux_installed() -> str | None:
    """
    Check if webpmux is installed.

    Returns:
        The version info of webpmux if it is installed, otherwise None.
    """

    try:
        # Attempt to run `webpmux -version` to check if webpmux is installed
        result = subprocess.run(["webpmux", "-version"], capture_output=True, text=True, check=True)
        # print("webpmux is installed. Version info:")
        # print(result.stdout)
        return result.stdout
    except subprocess.CalledProcessError as e:
        # The command was found, but it exited with an error
        # print("webpmux command failed:", e)
        return None
    except FileNotFoundError:
        # The command was not found
        # print("webpmux is not installed.")
        return None


def is_webp_animated(file_path: str) -> bool:
    """
    Check if a WebP file is animated.

    Args:
        file_path: The path to the WebP file.

    Returns:
        True if the WebP file is animated, False otherwise.

    Raises:
        FfmpegMediaTypeError: If the webpmux command fails.
    """

    # Running the webpmux command to get information about the WebP file
    result = subprocess.run(["webpmux", "-info", file_path], capture_output=True, text=True)
    output = result.stdout

    # Check output for the presence of 'ANMF' chunk which indicates animation
    if "Number of frames" in output:
        return True
    else:
        return False


def extract_animated_webp_frame(uri: str) -> str:
    """
    Fix an animated webp file by extracting the first frame.

    Args:
        uri: The URI of the webp file.

    Returns:
        The URI of the fixed webp file.

    Raises:
        FfmpegMediaTypeError: If the webpmux command fails.

    Notes:
        Will raise Exception if the webp file is not animated.
    """
    temp_uri = create_temp_filename(".webp")
    webpmux_command = ["webpmux", "-get", "frame", "1", uri, "-o", temp_uri]
    call(webpmux_command)
    return temp_uri


def is_webp_need_fix(uri: str | Path) -> bool:
    """
    Check if a WebP file needs to be fixed.

    Args:
        uri: The URI of the WebP file.

    Returns:
        True if the WebP file is animated, False otherwise.

    Raises:
        FfmpegMediaTypeError: If the ffprobe command fails.
    """

    ffprobe_cmd = ["ffprobe"] + [
        "-v",
        "error",
        "-show_format",
        "-show_streams",
        "-of",
        "json",
        str(uri),
    ]

    # Execute the FFprobe command and capture the output
    output = call(ffprobe_cmd)
    probe_info = from_dict(FFProbeInfo, json.loads(output))

    if probe_info.format.format_name != "webp_pipe":
        return False

    if probe_info.streams[0].height == 0 and probe_info.streams[0].width == 0:
        return True

    return False


def hotfix_animate_webp(uri: str | Path) -> str:
    """
    Fix a WebP file if it is animated.

    Args:
        uri: The URI of the WebP file.

    Returns:
        The URI of the fixed WebP file if it was animated, otherwise the original URI.

    Raises:
        FfmpegMediaTypeError: If the webpmux command fails.

    Notes:
        Some WebP files (e.g. animated WebP files) may not be supported by ffmpeg.
        This function can be used to fix such files by extracting the first frame.
        If webpmux is not installed, the original URI will be returned.
    """

    if check_webpmux_installed() is None:
        # webpmux is not installed, so we cannot fix the WebP file
        return str(uri)

    if not is_webp_need_fix(str(uri)):
        # The WebP file work with ffprobe, so no fix is needed
        return str(uri)

    downloaded_file = _ensure_downloaded(str(uri))
    if not is_webp_animated(downloaded_file):
        # The WebP file is not animated, so no fix is needed
        return str(uri)

    # The WebP file is animated, so extract the first frame
    return extract_animated_webp_frame(downloaded_file)



from .info import detect
from .schema import FFProbeInfo, MediaInfo
from .utils.ffprobe import ffprobe
from .utils.thumbnail import generate_thumbnail

__all__ = ["ffprobe", "MediaInfo", "detect", "FFProbeInfo", "generate_thumbnail"]



from pathlib import Path

from .hotfix_webp import hotfix_animate_webp
from .shell import call, create_temp_filename


def generate_thumbnail(video_path: str | Path, suffix: str = ".png", time_offset: float = 0) -> str:
    """
    Generate a thumbnail from a video file at a specified time offset.

    Args:
        video_path: the path to the video file
        suffix: the suffix of the generated thumbnail
        time_offset: the time offset in seconds to generate the thumbnail

    Raises:
        FfmpegMediaTypeError: If the ffmpeg command fails.

    Returns:
        the path to the generated thumbnail
    """

    video_path = hotfix_animate_webp(video_path)

    thumbnail_path = create_temp_filename(suffix)
    ffmpeg_cmd = ["ffmpeg"] + [
        "-y",  # Overwrite output file if it exists
        "-i",
        str(video_path),  # Input video path
        "-ss",
        str(time_offset),  # Time offset (seek to the specified position)
        "-vframes",
        "1",  # Number of frames to output
        "-vf",
        "scale=320:-1",  # Thumbnail size (width: 320, height: proportional)
        "-q:v",
        "2",  # Quality (2 - high, 5 - low)
        thumbnail_path,  # Output thumbnail path
    ]

    call(ffmpeg_cmd)

    return thumbnail_path



from dataclasses import dataclass, field
from typing import Literal


@dataclass(frozen=True, kw_only=True)
class FFMpegSupport:
    """
    FFMpeg support info for a codec.
    """

    demuxing_support: bool
    """
    Is ffmpeg support demuxing for the codec.
    """
    muxing_support: bool
    """
    Is ffmpeg support muxing for the codec.
    """
    codec: str
    """
    Codec name.
    """
    description: str
    """
    Codec description.
    """

    common_exts: tuple[str, ...]
    """
    Common file extensions for the codec.
    """
    mime_type: str
    """
    Mime type for the codec.
    """
    default_video_codec: str
    """
    Default video codec for the codec.
    """
    default_audio_codec: str
    """
    Default audio codec for the codec.
    """


@dataclass(frozen=True, kw_only=True)
class MediaInfo:
    """
    The Basic Media info.
    """

    type: Literal["image", "video", "audio"]
    """
    The media type.
    """

    width: int | None = None
    """
    The media width.
    """
    height: int | None = None
    """
    The media height.
    """
    duration: float | None = None
    """
    The media duration.
    """

    format: str | None = None
    """
    The media format.
    """
    size: int | None = None
    """
    The media size.
    """
    suggest_ext: str | None = None
    """
    The suggested file extension.
    """


@dataclass(kw_only=True, frozen=True)
class FFProbeFormat:
    """
    The media format info return by ffprobe.
    """

    filename: str | None = None
    """
    The media filename.
    """
    duration: str | None = None
    """
    The media duration.
    """
    format_name: str
    """
    The media format name.
    """
    format_long_name: str | None = None
    """
    The media format long name.
    """
    start_time: float | None = None
    """
    The media start time.
    """
    size: str | None = None
    """
    The media size.
    """
    probe_score: int | None = None
    """
    The media probe score.
    """


@dataclass(kw_only=True, frozen=True)
class FFProbeStreamTags:
    """
    The media stream tags return by ffprobe.
    """

    rotate: int = 0
    """
    The stream rotate value.
    """


@dataclass(kw_only=True, frozen=True)
class FFProbeStream:
    """
    The media stream info return by ffprobe.
    """

    index: int | None = None
    """
    The stream index.
    """
    width: int | None = None
    """
    The stream width.
    """
    height: int | None = None
    """
    The stream height.
    """
    codec_type: str | None = None
    """
    The stream codec type.
    """
    codec_name: str
    """
    The stream codec name.
    """
    codec_long_name: str | None = None
    """
    The stream codec long name.
    """
    profile: str | None = None
    """
    The stream profile.
    """
    pix_fmt: str | None = None
    """
    The stream pixel format.
    """
    r_frame_rate: str | None = None
    """
    The stream frame rate.
    """
    tags: FFProbeStreamTags = field(default_factory=FFProbeStreamTags)
    """
    The stream tags.
    """


@dataclass(kw_only=True, frozen=True)
class FFProbeInfo:
    """
    The media information return by ffprobe.
    """

    format: FFProbeFormat
    """
    The media format info.
    """
    streams: tuple[FFProbeStream, ...]
    """
    The media streams info.
    """



import subprocess
from shlex import join
from tempfile import NamedTemporaryFile

from ..exceptions import FfmpegMediaTypeError


def call(cmds: list[str]) -> str:
    """
    Run a command and return the stdout.

    Args:
        cmds: List of command and arguments.

    Raises:
        FfmpegMediaTypeError: If the command fails.

    Returns:
        stdout of the command.
    """
    try:
        r = subprocess.run(cmds, stdout=subprocess.PIPE, check=True)
    except subprocess.CalledProcessError as e:
        raise FfmpegMediaTypeError(f"command failed: {join(e.cmd)}")

    return r.stdout.decode("utf-8")


def create_temp_filename(suffix: str) -> str:
    """
    Create a temporary file path.

    Returns:
        the path to the temporary file
    """

    if not suffix.startswith("."):
        suffix = f".{suffix}"

    # Create a NamedTemporaryFile
    temp = NamedTemporaryFile(delete=True, suffix=suffix)
    temp_path = temp.name

    # Close and delete the temporary file (it won't actually create it)
    temp.close()

    # At this point, temp_path is a path to a non-existent file
    return temp_path



import json
from dataclasses import asdict, is_dataclass
from pathlib import Path
from typing import TypeVar

cache_path = Path(__file__).parent / "../cache"
cache_path.mkdir(exist_ok=True)

T = TypeVar("T")


def load(cls: type[T], name: str) -> T | None:
    """
    Load a dataclass object from a JSON file.

    Args:
        cls: The dataclass type.
        name: The name of the file.

    Returns:
        The dataclass object or None if the file does not exist.
    """
    path = cache_path / f"{name}.json"

    if not path.exists():
        return None

    with path.open() as ifile:
        return cls(**json.loads(ifile.read()))


def save(obj: T, name: str) -> None:
    """
    Save a dataclass object to a JSON file.

    Args:
        obj: The dataclass object.
        name: The name of the file.

    Returns:
        None
    """
    assert is_dataclass(obj) and not isinstance(obj, type)

    with (cache_path / f"{name}.json").open("w") as ofile:
        ofile.write(json.dumps(asdict(obj)))



from dataclasses import fields, is_dataclass
from typing import Any, TypeVar, get_args, get_origin

T = TypeVar("T")


def from_dict(data_class: type[T], data: dict[str, Any]) -> T:
    """
    Convert a dictionary to a dataclass object.

    Args:
        data_class: The dataclass type.
        data: The dictionary to convert.

    Returns:
        The dataclass object.
    """

    if isinstance(data, dict):
        assert is_dataclass(data_class)
        fieldtypes = {f.name: f.type for f in fields(data_class)}
        return data_class(**{f: from_dict(fieldtypes[f], data[f]) for f in data if f in fieldtypes})  # type: ignore[return-value]
    elif get_origin(data_class) == tuple:
        # Handling tuple types by converting each element in the tuple
        # assuming all elements are of the type specified in the first argument of the tuple type
        element_type = get_args(data_class)[0]
        return tuple(from_dict(element_type, item) for item in data)
    else:
        return data



import os
from urllib.parse import urlparse

from .schema import FFMpegSupport, MediaInfo
from .utils.cache import load
from .utils.ffprobe import ffprobe
from .utils.thumbnail import generate_thumbnail


def extract_file_extension_from_uri(uri: str) -> str:
    """
    Extract the file extension from a URI.

    Args:
        uri: the URI to extract the file extension from

    Returns:
        the file extension
    """

    parsed_uri = urlparse(uri)
    path = parsed_uri.path
    filename = os.path.basename(path)
    _, extension = os.path.splitext(filename)
    return extension[1:].lower()


KNOWN_CODEC_EXTS = {
    "png_pipe": ("png",),
    "svg_pipe": ("svg",),
    "tiff_pipe": ("tiff", "tif"),
    "bmp_pipe": ("bmp",),
    "gif_pipe": ("gif",),
    "jpeg_pipe": ("jpeg", "jpg"),
    "webp_pipe": ("webp",),
    "mjpeg": ("jpg", "jpeg", "mjpeg"),
}
"""
Dictionary mapping codec names to common file extensions.
"""


def detect(uri: str) -> MediaInfo:
    """
    Detect the media type of a file.

    Args:
        uri: the URI of the file

    Returns:
        the media type information

    Raises:
        FfmpegMediaTypeError: If the ffmpeg command fails.
    """
    info = ffprobe(uri)
    current_ext = extract_file_extension_from_uri(uri)

    format_name = info.format.format_name
    duration = info.format.duration

    # NOTE: handle ffmpeg's image compatibility
    if format_name == "image2":
        format_name = info.streams[0].codec_name

    # NOTE: detect file extension
    if format_name in KNOWN_CODEC_EXTS:
        common_exts = KNOWN_CODEC_EXTS[format_name]
    elif support_info := load(FFMpegSupport, format_name):
        common_exts = support_info.common_exts
    else:
        common_exts = ()

    if current_ext in common_exts:
        suggest_ext = current_ext
    elif common_exts:
        suggest_ext = common_exts[0]
    else:
        suggest_ext = None

    # NOTE: we classify gif as image
    if not duration or format_name in ("gif", "mjpeg"):
        return MediaInfo(
            type="image",
            width=info.streams[0].width or 0,
            height=info.streams[0].height or 0,
            duration=float(duration) if duration is not None else None,
            format=format_name,
            size=int(info.format.size) if info.format.size is not None else None,
            suggest_ext=suggest_ext,
        )

    for stream in info.streams:
        # NOTE: if there is at least one video stream, the media is video
        if stream.codec_type == "video" and format_name not in ("mp3",):
            width = stream.width
            height = stream.height

            return MediaInfo(
                type="video",
                width=width or 0,
                height=height or 0,
                duration=float(duration) if duration is not None else None,
                format=format_name,
                size=int(info.format.size) if info.format.size is not None else None,
                suggest_ext=suggest_ext,
            )

    # NOTE: if there is no video stream, the media is audio
    return MediaInfo(
        type="audio",
        width=0,
        height=0,
        duration=float(duration) if duration is not None else None,
        format=format_name,
        size=int(info.format.size) if info.format.size is not None else None,
        suggest_ext=suggest_ext,
    )


__all__ = ["detect", "generate_thumbnail"]
